from typing import Callable, NamedTuple

import jsonpickle
from jetpack.proto import remote_api_pb2


def encode_result(result) -> remote_api_pb2.RemoteCallResponse:
    response = remote_api_pb2.RemoteCallResponse()
    response.json_results = jsonpickle.encode(result)
    return response


def decode_result(response: remote_api_pb2.RemoteCallResponse):
    return jsonpickle.decode(response.json_results)


# FuncArgs is used by jsonpickle to capture the arguments to an RPC
# TODO: Consider using inspect.BoundArguments instead of defining our own type.
FuncArgs = NamedTuple("FuncArgs", [("args", tuple), ("kwargs", dict)])


def encode_args(args, kwargs) -> str:
    return jsonpickle.encode(FuncArgs(args=args, kwargs=kwargs))


def decode_args(encoded_args: str):
    func_args: FuncArgs = jsonpickle.decode(encoded_args)
    return func_args.args, func_args.kwargs



import json
import sys

import redis
import schedule
from google.protobuf import json_format
from jetpack import utils
from jetpack._job.job import Job
from jetpack.models.runtime import cronjob_pb2, describe_pb2, job_pb2


def describe_output(app):
    # Should these be the same? The concepts are really similar.
    cron_jobs = []
    jobs = []
    for job in schedule.get_jobs():

        if job.at_time is not None:
            target_time = job.at_time.isoformat()
        else:
            target_time = None

        if job.start_day is not None:
            target_day_of_week = cronjob_pb2.DayOfWeek.Value(job.start_day.upper())
        else:
            target_day_of_week = None

        cron_jobs.append(
            cronjob_pb2.CronJob(
                function=utils.job_name(job),
                target_time=target_time,
                target_day_of_week=target_day_of_week,
                unit=cronjob_pb2.Unit.Value(job.unit.upper()),
                interval=job.interval,
            )
        )

    for job in Job.defined_jobs():
        jobs.append(
            job_pb2.Job(
                qualified_symbol=job.name(),
            )
        )

    try:
        enum_name = app.__class__.__name__.upper() if app is not None else "NONE"
        framework = describe_pb2.Framework.Value(enum_name)
    except ValueError:
        framework = describe_pb2.Framework.Value("UNKNOWN")

    return describe_pb2.DescribeOutput(
        cron_jobs=cron_jobs,
        jobs=jobs,
        framework=framework,
    )


# TODO(Landau): Use a framework? Like https://click.palletsprojects.com/en/7.x/
def handle(app=None):
    """
    Call individual function: `python jetpack_main.py run func_name`
    """
    if (len(sys.argv) == 3 or len(sys.argv) == 4) and sys.argv[1] == "run":
        target_job_name = sys.argv[2]
        target_job = Job.find_job(target_job_name)
        if target_job:
            if len(sys.argv) == 4:
                target_job.exec(sys.argv[3])
            else:
                target_job.exec()
            return

        for job in schedule.get_jobs():
            if utils.job_name(job) == target_job_name:
                job.job_func()

    """
    Get all jobs `python jetpack_main.py describe`
    """
    if len(sys.argv) == 2 and sys.argv[1] == "describe":
        print(json_format.MessageToJson(describe_output(app)))

    """
    Get all jobs `python jetpack_main.py describe-to-redis`
    """
    if len(sys.argv) == 4 and sys.argv[1] == "describe-to-redis":
        host = sys.argv[2]
        key = sys.argv[3]
        r = redis.Redis(host=host, port=6379, db=0)
        r.set(key, json_format.MessageToJson(describe_output(app)))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sdk/proto/remote_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sdk/proto/remote_api.proto',
  package='remote',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1asdk/proto/remote_api.proto\x12\x06remote\"@\n\x11RemoteCallRequest\x12\x18\n\x10qualified_symbol\x18\x01 \x01(\t\x12\x11\n\tjson_args\x18\x02 \x01(\t\"*\n\x12RemoteCallResponse\x12\x14\n\x0cjson_results\x18\x01 \x01(\t\"2\n\x10LaunchJobRequest\x12\x1e\n\x03job\x18\x01 \x01(\x0b\x32\x11.remote.RemoteJob\"\x13\n\x11LaunchJobResponse\"Q\n\tRemoteJob\x12\x17\n\x0f\x63ontainer_image\x18\x01 \x01(\t\x12\x18\n\x10qualified_symbol\x18\x02 \x01(\t\x12\x11\n\tnamespace\x18\x03 \x01(\t2\x9b\x01\n\x0eRemoteExecutor\x12\x45\n\nRemoteCall\x12\x19.remote.RemoteCallRequest\x1a\x1a.remote.RemoteCallResponse\"\x00\x12\x42\n\tLaunchJob\x12\x18.remote.LaunchJobRequest\x1a\x19.remote.LaunchJobResponse\"\x00\x62\x06proto3'
)




_REMOTECALLREQUEST = _descriptor.Descriptor(
  name='RemoteCallRequest',
  full_name='remote.RemoteCallRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='qualified_symbol', full_name='remote.RemoteCallRequest.qualified_symbol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='json_args', full_name='remote.RemoteCallRequest.json_args', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=102,
)


_REMOTECALLRESPONSE = _descriptor.Descriptor(
  name='RemoteCallResponse',
  full_name='remote.RemoteCallResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='json_results', full_name='remote.RemoteCallResponse.json_results', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=146,
)


_LAUNCHJOBREQUEST = _descriptor.Descriptor(
  name='LaunchJobRequest',
  full_name='remote.LaunchJobRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='job', full_name='remote.LaunchJobRequest.job', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=198,
)


_LAUNCHJOBRESPONSE = _descriptor.Descriptor(
  name='LaunchJobResponse',
  full_name='remote.LaunchJobResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=200,
  serialized_end=219,
)


_REMOTEJOB = _descriptor.Descriptor(
  name='RemoteJob',
  full_name='remote.RemoteJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='container_image', full_name='remote.RemoteJob.container_image', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qualified_symbol', full_name='remote.RemoteJob.qualified_symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='remote.RemoteJob.namespace', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=221,
  serialized_end=302,
)

_LAUNCHJOBREQUEST.fields_by_name['job'].message_type = _REMOTEJOB
DESCRIPTOR.message_types_by_name['RemoteCallRequest'] = _REMOTECALLREQUEST
DESCRIPTOR.message_types_by_name['RemoteCallResponse'] = _REMOTECALLRESPONSE
DESCRIPTOR.message_types_by_name['LaunchJobRequest'] = _LAUNCHJOBREQUEST
DESCRIPTOR.message_types_by_name['LaunchJobResponse'] = _LAUNCHJOBRESPONSE
DESCRIPTOR.message_types_by_name['RemoteJob'] = _REMOTEJOB
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RemoteCallRequest = _reflection.GeneratedProtocolMessageType('RemoteCallRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOTECALLREQUEST,
  '__module__' : 'sdk.proto.remote_api_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteCallRequest)
  })
_sym_db.RegisterMessage(RemoteCallRequest)

RemoteCallResponse = _reflection.GeneratedProtocolMessageType('RemoteCallResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOTECALLRESPONSE,
  '__module__' : 'sdk.proto.remote_api_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteCallResponse)
  })
_sym_db.RegisterMessage(RemoteCallResponse)

LaunchJobRequest = _reflection.GeneratedProtocolMessageType('LaunchJobRequest', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHJOBREQUEST,
  '__module__' : 'sdk.proto.remote_api_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchJobRequest)
  })
_sym_db.RegisterMessage(LaunchJobRequest)

LaunchJobResponse = _reflection.GeneratedProtocolMessageType('LaunchJobResponse', (_message.Message,), {
  'DESCRIPTOR' : _LAUNCHJOBRESPONSE,
  '__module__' : 'sdk.proto.remote_api_pb2'
  # @@protoc_insertion_point(class_scope:remote.LaunchJobResponse)
  })
_sym_db.RegisterMessage(LaunchJobResponse)

RemoteJob = _reflection.GeneratedProtocolMessageType('RemoteJob', (_message.Message,), {
  'DESCRIPTOR' : _REMOTEJOB,
  '__module__' : 'sdk.proto.remote_api_pb2'
  # @@protoc_insertion_point(class_scope:remote.RemoteJob)
  })
_sym_db.RegisterMessage(RemoteJob)



_REMOTEEXECUTOR = _descriptor.ServiceDescriptor(
  name='RemoteExecutor',
  full_name='remote.RemoteExecutor',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=305,
  serialized_end=460,
  methods=[
  _descriptor.MethodDescriptor(
    name='RemoteCall',
    full_name='remote.RemoteExecutor.RemoteCall',
    index=0,
    containing_service=None,
    input_type=_REMOTECALLREQUEST,
    output_type=_REMOTECALLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LaunchJob',
    full_name='remote.RemoteExecutor.LaunchJob',
    index=1,
    containing_service=None,
    input_type=_LAUNCHJOBREQUEST,
    output_type=_LAUNCHJOBRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_REMOTEEXECUTOR)

DESCRIPTOR.services_by_name['RemoteExecutor'] = _REMOTEEXECUTOR

# @@protoc_insertion_point(module_scope)



from typing import Callable

import schedule


def job_name(job: schedule.Job) -> str:
    return qualified_func_name(job.job_func.func)


def qualified_func_name(func: Callable) -> str:
    return f"{func.__module__}.{func.__name__}"



import json
from typing import Dict, Optional


class NamespaceCache:
    def __init__(self):
        self.cache = None

    def __call__(self) -> str:
        if not self.cache:
            with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as f:
                self.cache = f.read().strip()
        return self.cache


get = NamespaceCache()



import json
from typing import Dict, Optional


class ModuleNotFoundError(Exception):
    pass


class SymbolMapCache:
    def __init__(self):
        self.cache = None

    def __call__(self) -> Dict[str, str]:
        if not self.cache:
            with open("/var/run/config/jetpack.io/symbol-to-container.json") as f:
                self.cache = json.loads(f.read())
        return self.cache

    def find_image_for_module(self, module: str) -> Optional[str]:
        symbol_map = self()
        if module not in symbol_map:
            raise ModuleNotFoundError("Image for module not fount")
        return symbol_map[module]


_symbol_map_cache = SymbolMapCache()
find_image_for_module = _symbol_map_cache.find_image_for_module



import os
from typing import Callable

import grpc
from jetpack._remote import codec
from jetpack.config import namespace, symbols
from jetpack.proto import remote_api_pb2, remote_api_pb2_grpc


class NoControllerAddressError(Exception):
    pass


class Client:
    def __init__(self):
        host = os.environ.get(
            "JETPACK_CONTROLLER_HOST", "compiler-service.jetpack.svc.cluster.local"
        )
        port = os.environ.get("JETPACK_CONTROLLER_PORT", "8080")
        self.address = f"{host}:{port}"
        self.channel = None
        self.stub = None
        self.is_dialed = False  # TODO: Mutex needed?
        self.dry_run = False

    def dial(self):
        if not self.address:
            raise NoControllerAddressError("Controller address is not set")
        # Since this is inter-cluster communication, insecure is fine.
        # In the future this won't even leave the pod, and use a sidecar so
        # it will be localhost.
        self.channel = grpc.insecure_channel(self.address)
        self.stub = remote_api_pb2_grpc.RemoteExecutorStub(self.channel)
        self.is_dialed = True

    def launch_job(self, qualified_name: str, module: str):
        """Launches a k8s job. For now this function assumes job will live in
        same namespace where the launcher is located.

        Keyword arguments:
        qualified_name -- qualified name as produced by utils.qualified_func_name
        module -- jetpack module where the job resides. Used to determine correct
        docker image.
        """
        job = remote_api_pb2.RemoteJob(
            container_image=symbols.find_image_for_module(module),
            qualified_symbol=qualified_name,
            namespace=namespace.get(),
        )
        request = remote_api_pb2.LaunchJobRequest(job=job)

        if self.dry_run:
            print("Dry Run:", request)
            return

        # If dialing is slow, consider dialing earlier.
        if not self.is_dialed:
            self.dial()
        return self.stub.LaunchJob(request)



from concurrent import futures
from typing import Callable, NamedTuple

import grpc
import jsonpickle
from jetpack._remote import codec
from jetpack.proto import remote_api_pb2, remote_api_pb2_grpc

# FuncArgs is used by jsonpickle to capture the arguments to a RPC
FuncArgs = NamedTuple("FuncArgs", [("args", tuple), ("kwargs", dict)])


class Servicer(remote_api_pb2_grpc.RemoteExecutorServicer):
    def __init__(self):
        # TODO: Figure out if we need any locking around the symbol table.
        self.symbol_table = {}  # TODO: lock?

    def export(self, fn):
        symbol = fn.__name__
        self.symbol_table[symbol] = fn
        return symbol

    def RemoteCall(self, request, context):
        fn = self.symbol_table[request.method_name]
        args, kwargs = codec.decode_args(request.json_args)
        result = fn(*args, **kwargs)
        return codec.encode_result(result)


class Server:
    def __init__(self):
        self.servicer = Servicer()
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        remote_api_pb2_grpc.add_RemoteExecutorServicer_to_server(
            self.servicer, self.server
        )
        self.is_listening = False  # TODO: Mutex needed?

    def Listen(self):
        self.server.add_insecure_port("[::]:50051")
        self.server.start()
        self.is_listening = True

    def export(self, fn):
        # Connect to the network lazily
        if not self.is_listening:
            self.Listen()
        return self.servicer.export(fn)



import functools
from typing import Callable, NamedTuple

import jsonpickle
from jetpack import _remote
from jetpack._remote import codec
from jetpack.proto import remote_api_pb2

SERVER = _remote.Server()
CLIENT = _remote.Client()


def remote(fn: Callable) -> Callable:  # Decorator
    symbol = export(fn)
    stub = remote_function(symbol)
    return functools.wraps(fn)(stub)


def export(fn: Callable) -> str:
    return SERVER.export(fn)


def remote_function(symbol: str) -> Callable:
    return CLIENT.remote_function(symbol)



from typing import Callable

from jetpack import _job
from jetpack._job.job import Job as _Job

_client = _job.Client()


class JobDecorator:
    def __call__(self, fn: Callable) -> Callable:
        _Job(fn)
        return fn

    @classmethod
    def launch(cls, qualified_name: str, module=""):
        _client.launch_job(qualified_name, module)


job = JobDecorator()



import os
from functools import wraps

import cronitor
import schedule
from jetpack import utils
from schedule import every  # Use this to whitelist what we allow

cronjob_suffix = os.environ.get("JETPACK_CRONJOB_SUFFIX", "-missing-suffix")


def repeat(repeat_pattern):
    def wrapper(func):
        job_name = utils.qualified_func_name(func) + cronjob_suffix
        cronitor_wrapped_func = cronitor.job(job_name)(func)
        schedule.repeat(repeat_pattern)(cronitor_wrapped_func)

    return wrapper



from typing import Callable

import grpc
from jetpack._remote import codec
from jetpack.proto import remote_api_pb2, remote_api_pb2_grpc


class Client:
    def __init__(self):
        self.address = "localhost:50051"
        self.channel = None
        self.stub = None
        self.is_dialed = False  # TODO: Mutex needed?

    def Dial(self):
        self.channel = grpc.insecure_channel(self.address)
        self.stub = remote_api_pb2_grpc.RemoteExecutorStub(self.channel)
        self.is_dialed = True

    def RemoteCall(self, request):
        # Lazily connect to the network. Putting this here so we get a clean
        # API from the get-go, but in practice we might want to do the network
        # connection when a function is *imported* as a stub, rather than on the
        # actual function call (to reduce latency on the call)
        if not self.is_dialed:
            self.Dial()
        return self.stub.RemoteCall(request)

    def remote_function(self, symbol: str) -> Callable:
        def rpc_wrapper(*args, **kwargs):
            request = call_as_proto(symbol, args, kwargs)
            response = self.RemoteCall(request)
            return codec.decode_result(response)

        return rpc_wrapper


def call_as_proto(symbol: str, args, kwargs) -> remote_api_pb2.RemoteCallRequest:
    request = remote_api_pb2.RemoteCallRequest()
    request.method_name = symbol
    request.json_args = codec.encode_args(args, kwargs)
    return request



from __future__ import annotations

from typing import Callable, Dict, Optional, ValuesView

from jetpack import utils
from jetpack._remote import codec


class AlreadyExistsError(Exception):
    pass


class Job:
    jobs: Dict[str, Job] = {}

    def __init__(self, func: Callable):
        self.func = func
        Job.add_job(self)

    def name(self) -> str:
        return utils.qualified_func_name(self.func)

    def exec(self, encoded_args: str = "") -> str:
        args = []
        kwargs = {}
        if encoded_args != "":
            args, kwargs = codec.decode_args(encoded_args)
        result = self.func(*args, **kwargs)
        # for now just return encoded result, but we could send it to a redis channel
        # that the caller might be listening to.
        return codec.encode_result(result)

    @classmethod
    def add_job(cls, job: Job):
        if job.name() in cls.jobs:
            raise AlreadyExistsError(f"Job with name {job.name()} already exists")
        cls.jobs[job.name()] = job

    @classmethod
    def find_job(cls, name: str) -> Optional[Job]:
        return cls.jobs.get(name)

    @classmethod
    def defined_jobs(cls) -> ValuesView[Job]:
        return cls.jobs.values()
